# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint/crypto/proof.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from packages.valory.connections.abci.gogoproto import (
    gogo_pb2 as gogoproto_dot_gogo__pb2,
)


DESCRIPTOR = _descriptor.FileDescriptor(
    name="tendermint/crypto/proof.proto",
    package="tendermint.crypto",
    syntax="proto3",
    serialized_options=b"Z8github.com/tendermint/tendermint/proto/tendermint/crypto",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1dtendermint/crypto/proof.proto\x12\x11tendermint.crypto\x1a\x14gogoproto/gogo.proto"G\n\x05Proof\x12\r\n\x05total\x18\x01 \x01(\x03\x12\r\n\x05index\x18\x02 \x01(\x03\x12\x11\n\tleaf_hash\x18\x03 \x01(\x0c\x12\r\n\x05\x61unts\x18\x04 \x03(\x0c"?\n\x07ValueOp\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\'\n\x05proof\x18\x02 \x01(\x0b\x32\x18.tendermint.crypto.Proof"6\n\x08\x44ominoOp\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05input\x18\x02 \x01(\t\x12\x0e\n\x06output\x18\x03 \x01(\t"2\n\x07ProofOp\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c"9\n\x08ProofOps\x12-\n\x03ops\x18\x01 \x03(\x0b\x32\x1a.tendermint.crypto.ProofOpB\x04\xc8\xde\x1f\x00\x42:Z8github.com/tendermint/tendermint/proto/tendermint/cryptob\x06proto3',
    dependencies=[
        gogoproto_dot_gogo__pb2.DESCRIPTOR,
    ],
)


_PROOF = _descriptor.Descriptor(
    name="Proof",
    full_name="tendermint.crypto.Proof",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="total",
            full_name="tendermint.crypto.Proof.total",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="index",
            full_name="tendermint.crypto.Proof.index",
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="leaf_hash",
            full_name="tendermint.crypto.Proof.leaf_hash",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="aunts",
            full_name="tendermint.crypto.Proof.aunts",
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=74,
    serialized_end=145,
)


_VALUEOP = _descriptor.Descriptor(
    name="ValueOp",
    full_name="tendermint.crypto.ValueOp",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="tendermint.crypto.ValueOp.key",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="proof",
            full_name="tendermint.crypto.ValueOp.proof",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=147,
    serialized_end=210,
)


_DOMINOOP = _descriptor.Descriptor(
    name="DominoOp",
    full_name="tendermint.crypto.DominoOp",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="tendermint.crypto.DominoOp.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="input",
            full_name="tendermint.crypto.DominoOp.input",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="output",
            full_name="tendermint.crypto.DominoOp.output",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=212,
    serialized_end=266,
)


_PROOFOP = _descriptor.Descriptor(
    name="ProofOp",
    full_name="tendermint.crypto.ProofOp",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="type",
            full_name="tendermint.crypto.ProofOp.type",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="key",
            full_name="tendermint.crypto.ProofOp.key",
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="tendermint.crypto.ProofOp.data",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=268,
    serialized_end=318,
)


_PROOFOPS = _descriptor.Descriptor(
    name="ProofOps",
    full_name="tendermint.crypto.ProofOps",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="ops",
            full_name="tendermint.crypto.ProofOps.ops",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=b"\310\336\037\000",
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=320,
    serialized_end=377,
)

_VALUEOP.fields_by_name["proof"].message_type = _PROOF
_PROOFOPS.fields_by_name["ops"].message_type = _PROOFOP
DESCRIPTOR.message_types_by_name["Proof"] = _PROOF
DESCRIPTOR.message_types_by_name["ValueOp"] = _VALUEOP
DESCRIPTOR.message_types_by_name["DominoOp"] = _DOMINOOP
DESCRIPTOR.message_types_by_name["ProofOp"] = _PROOFOP
DESCRIPTOR.message_types_by_name["ProofOps"] = _PROOFOPS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Proof = _reflection.GeneratedProtocolMessageType(
    "Proof",
    (_message.Message,),
    {
        "DESCRIPTOR": _PROOF,
        "__module__": "tendermint.crypto.proof_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.crypto.Proof)
    },
)
_sym_db.RegisterMessage(Proof)

ValueOp = _reflection.GeneratedProtocolMessageType(
    "ValueOp",
    (_message.Message,),
    {
        "DESCRIPTOR": _VALUEOP,
        "__module__": "tendermint.crypto.proof_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.crypto.ValueOp)
    },
)
_sym_db.RegisterMessage(ValueOp)

DominoOp = _reflection.GeneratedProtocolMessageType(
    "DominoOp",
    (_message.Message,),
    {
        "DESCRIPTOR": _DOMINOOP,
        "__module__": "tendermint.crypto.proof_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.crypto.DominoOp)
    },
)
_sym_db.RegisterMessage(DominoOp)

ProofOp = _reflection.GeneratedProtocolMessageType(
    "ProofOp",
    (_message.Message,),
    {
        "DESCRIPTOR": _PROOFOP,
        "__module__": "tendermint.crypto.proof_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.crypto.ProofOp)
    },
)
_sym_db.RegisterMessage(ProofOp)

ProofOps = _reflection.GeneratedProtocolMessageType(
    "ProofOps",
    (_message.Message,),
    {
        "DESCRIPTOR": _PROOFOPS,
        "__module__": "tendermint.crypto.proof_pb2"
        # @@protoc_insertion_point(class_scope:tendermint.crypto.ProofOps)
    },
)
_sym_db.RegisterMessage(ProofOps)


DESCRIPTOR._options = None
_PROOFOPS.fields_by_name["ops"]._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Contains the background tasks of the APY estimation skill."""

from typing import Any

import pandas as pd
from aea.skills.tasks import Task

from packages.valory.skills.apy_estimation_abci.ml.forecasting import (
    PoolIdToForecasterType,
    PoolIdToTestReportType,
    estimate_apy_per_pool,
    test_forecaster_per_pool,
    train_forecaster_per_pool,
    update_forecaster_per_pool,
)
from packages.valory.skills.apy_estimation_abci.ml.optimization import (
    PoolToHyperParamsWithStatusType,
    optimize,
)
from packages.valory.skills.apy_estimation_abci.ml.preprocessing import (
    TrainTestSplitType,
    prepare_pair_data,
)
from packages.valory.skills.apy_estimation_abci.tools.etl import (
    prepare_batch,
    transform_hist_data,
)


class TransformTask(Task):
    """Transform historical data."""

    def execute(self, *args: Any, **kwargs: Any) -> pd.DataFrame:
        """Execute the task."""
        return transform_hist_data(*args, **kwargs)


class PreprocessTask(Task):
    """Preprocess historical data."""

    def execute(self, *args: Any, **kwargs: Any) -> TrainTestSplitType:
        """Execute the task."""
        return prepare_pair_data(*args, **kwargs)


class PrepareBatchTask(Task):
    """Prepare a batch."""

    def execute(self, *args: Any, **kwargs: Any) -> pd.DataFrame:
        """Execute the task."""
        return prepare_batch(*args, **kwargs)


class OptimizeTask(Task):
    """Run an optimization study per pool."""

    def execute(self, *args: Any, **kwargs: Any) -> PoolToHyperParamsWithStatusType:
        """Execute the task."""
        return optimize(*args, **kwargs)


class TrainTask(Task):
    """Train forecasters."""

    def execute(self, *args: Any, **kwargs: Any) -> PoolIdToForecasterType:
        """Execute the task."""
        return train_forecaster_per_pool(*args, **kwargs)


class TestTask(Task):
    """Test forecasters."""

    def execute(self, *args: Any, **kwargs: Any) -> PoolIdToTestReportType:
        """Execute the task."""
        return test_forecaster_per_pool(*args, **kwargs)


class UpdateTask(Task):
    """Update forecasters."""

    def execute(self, *args: Any, **kwargs: Any) -> None:
        """Execute the task."""
        return update_forecaster_per_pool(*args, **kwargs)


class EstimateTask(Task):
    """Estimate APYs."""

    def execute(self, *args: Any, **kwargs: Any) -> pd.DataFrame:
        """Execute the task."""
        return estimate_apy_per_pool(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""
Read the full transaction history and show the ERC20 transfer events in a human-readable format for a set of known accounts.

To run with a Hardhat integration test, set a breakpoint before the test ends to avoid the local chain being destroyed and Hardhat stopped.
"""
import json
import sys
from enum import Enum
from typing import Optional, cast

from hexbytes import HexBytes
from web3 import Web3
from web3.logs import IGNORE


HARDHAT_ENDPOINT = "http://127.0.0.1:8545"
HARDHAT_DEFAULT_MNEMONIC = "test test test test test test test test test test test junk"
ERC20_ABI = "../third_party/contracts-amm/node_modules/@openzeppelin/contracts/build/contracts/ERC20PresetFixedSupply.json"


class Color(Enum):
    """Standard terminal color codes."""

    GREEN = "\033[32m"
    RESET = "\033[39m"


# These Hardhat addresses are assigned in the contracts-amm deployment.
hardhat_account_indices_to_names = {
    0: "token_A_issuer",
    1: "token_B_issuer",
    10: "safe_owner_1",
    11: "safe_owner_2",
    12: "safe_owner_3",
    13: "safe_owner_4",
}

# Known addresses from out Hardhat deployment
# It would be nice to have automatic derivation for these ones as we do for Hardhat default accounts.
account_names_to_addresses = {
    "gnosis_safe": "0x9fE46736679d2D9a65F0992F2272dE9f3c7fa6e0",
    "gnosis_safe_L2": "0xe7f1725E7734CE288F8367e1Bb143E90bb3F0512",
    "gnosis_safe_proxy_factory": "0x5FbDB2315678afecb367f032d93F642f64180aa3",
    "gnosis_multisend": "0x2279B7A0a67DB372996a5FaB50D91eAA73d2eBe6",
    "safe_proxy": "0x67C828DC58df857618c88994a4bA6c15f05ebf36",
    "uniswap_factory": "0xB7f8BC63BbcaD18155201308C8f3540b07f84F5e",
    "uniswap_router02": "0xA51c1fc2f0D1a1b8494Ed1FE312d7C3a78Ed91C0",
    "token_A": "0x0DCd1Bf9A1b36cE34237eEaFef220932846BCD82",
    "token_B": "0x9A676e781A523b5d0C0e43731313A708CB607508",
    "token_LP": "0x50cd56fb094f8f06063066a619d898475dd3eede",
    "token_WETH": "0xDc64a140Aa3E981100a9becA4E685f962f0cF6C9",
    "A_WETH_pool": "0x86A6C37D3E868580a65C723AAd7E0a945E170416",
    "B_WETH_pool": "0x3430fe46bfE23b1fafDe4F7c78481051F7c0E01F",
    "minter": "0x0000000000000000000000000000000000000000",
    "safe": "0x68FCdF52066CcE5612827E872c45767E5a1f6551",
    "defaultFallbackHandlerContractName": "0xCf7Ed3AccA5a467e9e704C703E8D87F634fB0Fc9",
    "SimulateTxAccessor": "0x5FC8d32690cc91D4c39d9d3abcBD16989F875707",
    "CompatibilityFallbackHandler": "0x0165878A594ca255338adfa4d48449f69242Eb8F",
    "CreateCall": "0xa513E6E4b8f2a923D98304ec87F64353C4D5C853",
    "MultiSendCallOnly": "0x8A791620dd6260079BF849Dc5567aDC3F2FdC318",
    "SignMessageLib": "0x610178dA211FEF7D417bC0e6FeD39F05609AD788",
}


def get_tx_tag(tx_logs: dict) -> str:
    """Identifies known transactions as ENTER_POOL, EXIT_POOL or SWAP_BACK.

    :param tx_logs: the transfer logs
    :return: the transaction tag
    """
    for log in tx_logs:
        if "errors" not in log:

            # If a tx has a LP_token transfer from the minter to the safe we flag it as ENTER_POOL
            if (
                log["args"]["from"] == account_names_to_addresses["minter"]
                and log["args"]["to"] == account_names_to_addresses["safe"]
                and log["address"] == account_names_to_addresses["token_LP"]
            ):
                return (
                    Color.GREEN.value + "ENTER_POOL" + Color.RESET.value
                )  # Using standard colored output

            # If a tx has a LP_token transfer from the safe to the LP pool we flag it as EXIT_POOL
            if (
                log["args"]["from"] == account_names_to_addresses["safe"]
                and log["args"]["to"] == account_names_to_addresses["token_LP"]
                and log["address"] == account_names_to_addresses["token_LP"]
            ):
                return (
                    Color.GREEN.value + "EXIT_POOL" + Color.RESET.value
                )  # Using standard colored output

            # If a tx has a token_A transfer from the pool to the safe we flag it as SWAP_BACK
            if (
                log["args"]["from"] == account_names_to_addresses["safe"]
                and log["args"]["to"] == account_names_to_addresses["A_WETH_pool"]
                and log["address"] == account_names_to_addresses["token_A"]
            ):
                return (
                    Color.GREEN.value + "SWAP_BACK" + Color.RESET.value
                )  # Using standard colored output

    return "N/A"


def get_address_name(address: Optional[str]) -> str:
    """Return the address name for known accounts.

    :param address: the address
    :return: the aaddress name
    """
    if address is None:
        return "none"
    address = cast(str, Web3.toChecksumAddress(address))
    return (
        addresses_to_account_names[address]
        if address in addresses_to_account_names
        else address
    )


if __name__ == "__main__":

    # Connect to HardHat
    w3 = Web3(Web3.HTTPProvider(HARDHAT_ENDPOINT))

    if not w3.isConnected():
        print("Not connected!")
        sys.exit()

    # Enable features to derive addresses
    w3.eth.account.enable_unaudited_hdwallet_features()

    # Checksum the addresses
    for acount_name in account_names_to_addresses.keys():
        account_names_to_addresses[acount_name] = Web3.toChecksumAddress(
            account_names_to_addresses[acount_name]
        )

    # Add the 20 Hardhat default accounts to known addresses
    for i in range(20):
        account = w3.eth.account.from_mnemonic(
            HARDHAT_DEFAULT_MNEMONIC, account_path=f"m/44'/60'/0'/0/{i}"
        )
        acount_name = (
            hardhat_account_indices_to_names[i]
            if i in hardhat_account_indices_to_names
            else f"hardhat_address_{i}"
        )
        account_names_to_addresses[acount_name] = Web3.toChecksumAddress(
            account.address
        )

    # Create the dict address -> account name
    addresses_to_account_names = {v: k for k, v in account_names_to_addresses.items()}

    # Load the ERC20 contract
    with open(ERC20_ABI, encoding="utf-8") as abi_file:
        erc_20_token_abi = json.load(abi_file)["abi"]
        erc_20_contract = w3.eth.contract(
            address=w3.toChecksumAddress(account_names_to_addresses["token_A"]),
            abi=erc_20_token_abi,
        )

    # Get all transactions and receipts in all blocks
    last_block_number = w3.eth.get_block("latest")["number"]

    table_header = f"{'FROM':<18}      {'TO':<18}  {'TOKEN VALUE':<14} {'TOKEN':<16}"

    for i in range(last_block_number + 1):
        block = w3.eth.get_block(i)

        for tx_hash in block["transactions"]:
            tx = w3.eth.get_transaction(cast(HexBytes, tx_hash))
            tx_receipt = w3.eth.get_transaction_receipt(cast(HexBytes, tx_hash))

            if tx_receipt is None:
                raise ValueError

            # Get the ERC20 transfer logs
            transfer_logs = erc_20_contract.events.Transfer().processReceipt(
                tx_receipt, IGNORE
            )

            # Identify known transactions
            tx_tag = get_tx_tag(transfer_logs)

            # Print the transaction info
            from_address = get_address_name(tx["from"])  # type: ignore
            to_address = get_address_name(tx["to"])  # type: ignore

            print()
            print(f"{'-'*75} Transaction {tx_hash.hex()}")  # type: ignore
            print(
                f"{table_header} from: {from_address}  | to: {to_address}  |  ETH value: {tx.value}"  # type: ignore
            )
            print(f"{' '*75} nonce: {tx.nonce}  |  block: {i}  |  tag: {tx_tag}")  # type: ignore
            print(
                f"{' '*75} gas used: {tx_receipt['gasUsed']}  |  gas limit: {tx.gas}  |  gas price: {tx.gasPrice}"  # type: ignore
            )

            # Print the tranfer events
            for tx_log in transfer_logs:
                if "errors" not in tx_log:
                    from_address = get_address_name(tx_log["args"]["from"])
                    to_address = get_address_name(tx_log["args"]["to"])
                    token = get_address_name(tx_log["address"])
                    value = tx_log["args"]["value"]
                    print(
                        f"{from_address:<18} ->   {to_address:<18}: {value:<14} {token:<16}"
                    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This tool generates the API docs."""
import argparse
import re
import shutil
import subprocess  # nosec
import sys
from pathlib import Path
from typing import Set

from aea.configurations.base import ComponentType, PublicId
from aea.configurations.constants import PACKAGES


DOCS_DIR = Path("docs/")
API_DIR = DOCS_DIR / "api/"
AEA_DIR = Path("autonomy")
PACKAGES_DIR = Path(PACKAGES)
DEFAULT_PACKAGES = {
    (ComponentType.CONNECTION, "valory/abci:latest"),
    (ComponentType.CONTRACT, "valory/gnosis_safe:latest"),
    (ComponentType.CONTRACT, "valory/gnosis_safe_proxy_factory:latest"),
    (ComponentType.PROTOCOL, "valory/abci:latest"),
    (ComponentType.SKILL, "valory/abstract_abci:latest"),
    (ComponentType.SKILL, "valory/abstract_round_abci:latest"),
    (ComponentType.SKILL, "valory/oracle_abci:latest"),
    (ComponentType.SKILL, "valory/oracle_deployment_abci:latest"),
    (ComponentType.SKILL, "valory/price_estimation_abci:latest"),
    (ComponentType.SKILL, "valory/registration_abci:latest"),
    (ComponentType.SKILL, "valory/safe_deployment_abci:latest"),
    (ComponentType.SKILL, "valory/transaction_settlement_abci:latest"),
}

IGNORE_NAMES = {r"^__init__\.py$", r"^__version__\.py$", r"^py\.typed$", r"^.*_pb2.py$"}
IGNORE_PREFIXES: Set[Path] = set()


def check_working_tree_is_dirty() -> None:
    """Check if the current Git working tree is dirty."""
    print("Checking if there are any newly generated doc file...")
    result = subprocess.check_output(["git", "status"])  # nosec
    if len(result) > 0:
        files = [
            file.replace("modified:", "").strip()
            for file in result.decode("utf-8").split("\n")
            if "modified:" in file
        ]
        filterd_md_files = [
            file for file in files if file.endswith(".md") and file.startswith("docs")
        ]
        if len(filterd_md_files) > 0:
            print("\nNewly generated doc files:\n")
            print("\n".join(filterd_md_files))
            sys.exit(1)

    print("All good!")


def create_subdir(path: str) -> None:
    """
    Create a subdirectory.

    :param path: the directory path
    """
    directory = "/".join(path.split("/")[:-1])
    Path(directory).mkdir(parents=True, exist_ok=True)


def replace_underscores(text: str) -> str:
    """
    Replace escaped underscores in a text.

    :param text: the text
    :return: the processed text
    """
    text_a = text.replace("\\_\\_", "`__`")
    text_b = text_a.replace("\\_", "`_`")
    return text_b


def is_relative_to(path_one: Path, path_two: Path) -> bool:
    """Check if a path is relative to another path."""
    return str(path_one).startswith(str(path_two))


def is_not_dir(path: Path) -> bool:
    """Call p.is_dir() method and negate the result."""
    return not path.is_dir()


def should_skip(module_path: Path) -> bool:
    """Return true if the file should be skipped."""
    if any(re.search(pattern, module_path.name) for pattern in IGNORE_NAMES):
        print("Skipping, it's in ignore patterns")
        return True
    if module_path.suffix != ".py":
        print("Skipping, it's not a Python module.")
        return True
    if any(is_relative_to(module_path, prefix) for prefix in IGNORE_PREFIXES):
        print(f"Ignoring prefix {module_path}")
        return True
    return False


def _generate_apidocs_aea_modules() -> None:
    """Generate API docs for aea.* modules."""
    for module_path in filter(is_not_dir, Path(AEA_DIR).rglob("*")):
        print(f"Processing {module_path}... ", end="")
        if should_skip(module_path):
            continue
        parents = module_path.parts[:-1]
        parents_without_root = module_path.parts[1:-1]
        last = module_path.stem
        doc_file = API_DIR / Path(*parents_without_root) / f"{last}.md"
        dotted_path = ".".join(parents) + "." + last
        make_pydoc(dotted_path, doc_file)


def _generate_apidocs_packages() -> None:
    """Generate API docs for default packages."""
    for component_type, default_package in DEFAULT_PACKAGES:
        public_id = PublicId.from_str(default_package)
        author = public_id.author
        name = public_id.name
        type_plural = component_type.to_plural()
        package_dir = PACKAGES_DIR / author / type_plural / name
        for module_path in package_dir.rglob("*.py"):
            print(f"Processing {module_path}...", end="")
            if should_skip(module_path):
                continue
            suffix = Path(str(module_path.relative_to(package_dir))[:-3] + ".md")
            dotted_path = ".".join(module_path.parts)[:-3]
            doc_file = API_DIR / type_plural / name / suffix
            make_pydoc(dotted_path, doc_file)


def make_pydoc(dotted_path: str, dest_file: Path) -> None:
    """Make a PyDoc file."""
    print(
        f"Running with dotted path={dotted_path} and dest_file={dest_file}... ", end=""
    )
    try:
        api_doc_content = run_pydoc_markdown(dotted_path)
        dest_file.parent.mkdir(parents=True, exist_ok=True)
        dest_file.write_text(api_doc_content)
    except Exception as e:  # pylint: disable=broad-except,invalid-name
        print(f"Error: {str(e)}")
        return
    print("Done!")


def run_pydoc_markdown(module: str) -> str:
    """
    Run pydoc-markdown.

    :param module: the dotted path.
    :return: the PyDoc content (pre-processed).
    """
    pydoc = subprocess.Popen(  # nosec # pylint: disable=consider-using-with
        ["pydoc-markdown", "-m", module, "-I", "."], stdout=subprocess.PIPE
    )
    stdout, _ = pydoc.communicate()
    pydoc.wait()
    stdout_text = stdout.decode("utf-8")
    text = replace_underscores(stdout_text)
    return text


def generate_api_docs() -> None:
    """Generate the api docs."""
    shutil.rmtree(API_DIR, ignore_errors=True)
    API_DIR.mkdir()
    _generate_apidocs_packages()
    _generate_apidocs_aea_modules()


def install(package: str) -> int:
    """
    Install a PyPI package by calling pip.

    :param package: the package name and version specifier.
    :return: the return code.
    """
    return subprocess.check_call(  # nosec
        [sys.executable, "-m", "pip", "install", package]
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser("generate_api_docs")
    parser.add_argument(
        "--check-clean", action="store_true", help="Check if the working tree is clean."
    )
    arguments = parser.parse_args()

    res = shutil.which("pydoc-markdown")
    if res is None:
        install("pydoc-markdown==3.3.0")
        sys.exit(1)

    generate_api_docs()
    if arguments.check_clean:
        check_working_tree_is_dirty()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/service_registry contract."""

from pathlib import Path
from typing import Any, Dict, List, Tuple
from unittest import mock

import pytest
from aea_ledger_ethereum import EthereumCrypto
from web3 import Web3

from packages.valory.contracts.service_registry.contract import (
    PUBLIC_ID,
    ServiceRegistryContract,
)

from tests.conftest import ROOT_DIR
from tests.test_contracts.base import BaseGanacheContractTest


Address = str  # hex
ConfigHash = Tuple[bytes, int, int]
AgentParams = Tuple[int, int]

ADDRESS_ZERO = "0x0000000000000000000000000000000000000000"
AGENT_REGISTRY = "0xe7f1725E7734CE288F8367e1Bb143E90bb3F0512"
OWNER = "0x8626f6940E2eb28930eFb4CeF49B2d1F2C9C1199"
SERVICE_REGISTRY = "0xA51c1fc2f0D1a1b8494Ed1FE312d7C3a78Ed91C0"

NONCE = 0
CHAIN_ID = 31337


class BaseServiceRegistryContractTest(BaseGanacheContractTest):
    """Base class for Service Registry contract tests"""

    contract: ServiceRegistryContract
    ledger_identifier = EthereumCrypto.identifier
    contract_address = SERVICE_REGISTRY
    contract_directory = Path(
        ROOT_DIR, "packages", PUBLIC_ID.author, "contracts", PUBLIC_ID.name
    )

    GAS: int = 10 ** 10
    DEFAULT_MAX_FEE_PER_GAS: int = 10 ** 10
    DEFAULT_MAX_PRIORITY_FEE_PER_GAS: int = 10 ** 10
    eth_value: int = 0

    @classmethod
    def deployment_kwargs(cls) -> Dict[str, Any]:
        """Contract deployment kwargs"""

        _name: str = "TestServiceRegistry"
        _symbol: str = "OLA"
        _agentRegistry: Address = Web3.toChecksumAddress(AGENT_REGISTRY)

        return dict(
            _name=_name,
            _symbol=_symbol,
            _agentRegistry=_agentRegistry,
            gas=cls.GAS,
        )

    def get_dummy_service_args(self) -> List:
        """Create dummy service"""

        owner: Address = OWNER
        name: str = "service name"
        description: str = "service description"
        config_hash: ConfigHash = (
            b'UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU"',
            int("0x12", 16),
            int("0x20", 16),
        )
        agent_ids: List[int] = [1, 2]
        agent_params: List[AgentParams] = [(3, 1000), (4, 1000)]
        threshold: int = 0

        return [
            owner,
            name,
            description,
            config_hash,
            agent_ids,
            agent_params,
            threshold,
        ]


class TestServiceRegistryContract(BaseServiceRegistryContractTest):
    """Test Service Registry Contract"""

    def test_deployment(self) -> None:
        """Test deployment"""

        expected_keys = {
            "gas",
            "chainId",
            "value",
            "nonce",
            "maxFeePerGas",
            "maxPriorityFeePerGas",
            "data",
            "from",
        }

        tx = self.contract.get_deploy_transaction(
            ledger_api=self.ledger_api,
            deployer_address=str(self.deployer_crypto.address),
            **self.deployment_kwargs(),
        )

        assert isinstance(tx, dict)
        assert len(tx) == 8
        assert not expected_keys.symmetric_difference(tx)
        assert tx["data"].startswith("0x") and len(tx["data"]) > 2

    def test_verify_contract(self) -> None:
        """Run verify test. If abi file is updated tests + addresses need updating"""

        assert self.contract_address is not None
        result = self.contract.verify_contract(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
        )

        assert result["verified"] is True

    @pytest.mark.parametrize("service_id, expected", [(0, False), (1, True)])
    def test_exists(self, service_id: int, expected: int) -> None:
        """Test whether service id exists"""

        hex_str = "0x" + "0" * 63 + str(int(expected))

        with mock.patch.object(
            self.ledger_api.api.manager, "request_blocking", return_value=hex_str
        ):
            exists = self.contract.exists(
                self.ledger_api,
                self.contract_address,
                service_id,
            )

        assert exists is expected

    def test_get_service_info(self) -> None:
        """Test service info retrieval"""

        return_value = dict(
            owner=OWNER,
            name="service name",
            description="service description",
            config_hash=(b'UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU"', 18, 32),
            threshold=7,
            num_agent_ids=2,
            agent_ids=[1, 2],
            agent_params=[(3, 1000), (4, 1000)],
            num_agent_instances=0,
            agent_instances=[],
            multisig=ADDRESS_ZERO,
        )

        assert self.contract_address is not None

        with mock.patch.object(
            self.ledger_api,
            "contract_method_call",
            return_value=list(return_value.values()),
        ):
            result = self.contract.get_service_info(
                ledger_api=self.ledger_api,
                contract_address=self.contract_address,
                service_id=1,
            )

        assert result == return_value



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the payloads.py module of the skill."""

from packages.valory.skills.price_estimation_abci.payloads import (
    EstimatePayload,
    ObservationPayload,
    TransactionHashPayload,
)
from packages.valory.skills.price_estimation_abci.payloads import (
    TransactionType as PETransactionType,
)
from packages.valory.skills.safe_deployment_abci.payloads import (
    DeploySafePayload,
    RandomnessPayload,
    SelectKeeperPayload,
)
from packages.valory.skills.safe_deployment_abci.payloads import (
    TransactionType as DSTransactionType,
)


def test_select_keeper_payload() -> None:
    """Test `SelectKeeperPayload`."""

    payload = SelectKeeperPayload(sender="sender", keeper="keeper")

    assert payload.keeper == "keeper"
    assert payload.data == {"keeper": "keeper"}
    assert payload.transaction_type == DSTransactionType.SELECT_KEEPER


def test_deploy_safe_payload() -> None:
    """Test `DeploySafePayload`."""

    payload = DeploySafePayload(sender="sender", safe_contract_address="address")

    assert payload.safe_contract_address == "address"
    assert payload.data == {"safe_contract_address": "address"}
    assert payload.transaction_type == DSTransactionType.DEPLOY_SAFE


def test_observation_payload() -> None:
    """Test `ObservationPayload`."""

    payload = ObservationPayload(sender="sender", observation=1.0)

    assert payload.observation == 1.0
    assert payload.data == {"observation": 1.0}
    assert payload.transaction_type == PETransactionType.OBSERVATION


def test_estimate_payload() -> None:
    """Test `EstimatePayload`."""

    payload = EstimatePayload(sender="sender", estimate=1.0)

    assert payload.estimate == 1.0
    assert payload.data == {"estimate": 1.0}
    assert payload.transaction_type == PETransactionType.ESTIMATE


def test_transaction_hash_payload() -> None:
    """Test `TransactionHashPayload`."""

    payload = TransactionHashPayload(sender="sender", tx_hash="hash")

    assert payload.tx_hash == "hash"
    assert payload.data == {"tx_hash": "hash"}
    assert payload.transaction_type == PETransactionType.TX_HASH


def test_randomness_payload() -> None:
    """Test `RandomnessPayload`"""

    payload = RandomnessPayload(sender="sender", round_id=1, randomness="1", id_="id")

    assert payload.round_id == 1
    assert payload.randomness == "1"
    assert payload.id_ == "id"
    assert payload.data == {"round_id": 1, "randomness": "1"}

    assert payload.transaction_type == DSTransactionType.RANDOMNESS



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the payloads.py module of the skill."""

from packages.valory.skills.hello_world_abci.payloads import (
    PrintMessagePayload,
    RegistrationPayload,
    ResetPayload,
    SelectKeeperPayload,
    TransactionType,
)


def test_registration_payload() -> None:
    """Test `RegistrationPayload`."""

    payload = RegistrationPayload(sender="sender")

    assert payload.sender == "sender"
    assert payload.transaction_type == TransactionType.REGISTRATION


def test_print_message_payload() -> None:
    """Test `PrintMessagePayload`."""

    payload = PrintMessagePayload(sender="sender", message="message")

    assert payload.message == "message"
    assert payload.data == {"message": "message"}
    assert payload.transaction_type == TransactionType.PRINT_MESSAGE


def test_select_keeper_payload() -> None:
    """Test `SelectKeeperPayload`."""

    payload = SelectKeeperPayload(sender="sender", keeper="keeper")

    assert payload.keeper == "keeper"
    assert payload.data == {"keeper": "keeper"}
    assert payload.transaction_type == TransactionType.SELECT_KEEPER


def test_reset_payload() -> None:
    """Test `ResetPayload`"""

    payload = ResetPayload(sender="sender", period_count=1, id_="id")

    assert payload.period_count == 1
    assert payload.id_ == "id"
    assert payload.data == {"period_count": 1}
    assert hash(payload) == hash(tuple(sorted(payload.data.items())))

    assert str(payload.transaction_type) == str(TransactionType.RESET)
    assert payload.transaction_type == TransactionType.RESET



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ledger_api.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10ledger_api.proto\x12\x1c\x61\x65\x61.valory.ledger_api.v1_0_0"\xde\x15\n\x10LedgerApiMessage\x12V\n\x07\x62\x61lance\x18\x05 \x01(\x0b\x32\x43.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Balance_PerformativeH\x00\x12R\n\x05\x65rror\x18\x06 \x01(\x0b\x32\x41.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Error_PerformativeH\x00\x12^\n\x0bget_balance\x18\x07 \x01(\x0b\x32G.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Get_Balance_PerformativeH\x00\x12n\n\x13get_raw_transaction\x18\x08 \x01(\x0b\x32O.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Get_Raw_Transaction_PerformativeH\x00\x12Z\n\tget_state\x18\t \x01(\x0b\x32\x45.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Get_State_PerformativeH\x00\x12v\n\x17get_transaction_receipt\x18\n \x01(\x0b\x32S.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Get_Transaction_Receipt_PerformativeH\x00\x12\x66\n\x0fraw_transaction\x18\x0b \x01(\x0b\x32K.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Raw_Transaction_PerformativeH\x00\x12v\n\x17send_signed_transaction\x18\x0c \x01(\x0b\x32S.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Send_Signed_Transaction_PerformativeH\x00\x12R\n\x05state\x18\r \x01(\x0b\x32\x41.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.State_PerformativeH\x00\x12l\n\x12transaction_digest\x18\x0e \x01(\x0b\x32N.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Transaction_Digest_PerformativeH\x00\x12n\n\x13transaction_receipt\x18\x0f \x01(\x0b\x32O.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Transaction_Receipt_PerformativeH\x00\x1a\x18\n\x06Kwargs\x12\x0e\n\x06kwargs\x18\x01 \x01(\x0c\x1a)\n\x0eRawTransaction\x12\x17\n\x0fraw_transaction\x18\x01 \x01(\x0c\x1a/\n\x11SignedTransaction\x12\x1a\n\x12signed_transaction\x18\x01 \x01(\x0c\x1a\x16\n\x05State\x12\r\n\x05state\x18\x01 \x01(\x0c\x1a\x16\n\x05Terms\x12\r\n\x05terms\x18\x01 \x01(\x0c\x1a/\n\x11TransactionDigest\x12\x1a\n\x12transaction_digest\x18\x01 \x01(\x0c\x1a\x31\n\x12TransactionReceipt\x12\x1b\n\x13transaction_receipt\x18\x01 \x01(\x0c\x1a>\n\x18Get_Balance_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x1ag\n Get_Raw_Transaction_Performative\x12\x43\n\x05terms\x18\x01 \x01(\x0b\x32\x34.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Terms\x1a\x84\x01\n$Send_Signed_Transaction_Performative\x12\\\n\x12signed_transaction\x18\x01 \x01(\x0b\x32@.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.SignedTransaction\x1a\xf0\x01\n$Get_Transaction_Receipt_Performative\x12\\\n\x12transaction_digest\x18\x01 \x01(\x0b\x32@.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.TransactionDigest\x12\x15\n\rretry_timeout\x18\x02 \x01(\x05\x12\x1c\n\x14retry_timeout_is_set\x18\x03 \x01(\x08\x12\x16\n\x0eretry_attempts\x18\x04 \x01(\x05\x12\x1d\n\x15retry_attempts_is_set\x18\x05 \x01(\x08\x1a:\n\x14\x42\x61lance_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x05\x1av\n\x1cRaw_Transaction_Performative\x12V\n\x0fraw_transaction\x18\x01 \x01(\x0b\x32=.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.RawTransaction\x1a\x7f\n\x1fTransaction_Digest_Performative\x12\\\n\x12transaction_digest\x18\x01 \x01(\x0b\x32@.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.TransactionDigest\x1a\x82\x01\n Transaction_Receipt_Performative\x12^\n\x13transaction_receipt\x18\x01 \x01(\x0b\x32\x41.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.TransactionReceipt\x1a\x92\x01\n\x16Get_State_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x10\n\x08\x63\x61llable\x18\x02 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x03 \x03(\t\x12\x45\n\x06kwargs\x18\x04 \x01(\x0b\x32\x35.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Kwargs\x1al\n\x12State_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x43\n\x05state\x18\x02 \x01(\x0b\x32\x34.aea.valory.ledger_api.v1_0_0.LedgerApiMessage.State\x1an\n\x12\x45rror_Performative\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x16\n\x0emessage_is_set\x18\x03 \x01(\x08\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x13\n\x0b\x64\x61ta_is_set\x18\x05 \x01(\x08\x42\x0e\n\x0cperformativeb\x06proto3'
)


_LEDGERAPIMESSAGE = DESCRIPTOR.message_types_by_name["LedgerApiMessage"]
_LEDGERAPIMESSAGE_KWARGS = _LEDGERAPIMESSAGE.nested_types_by_name["Kwargs"]
_LEDGERAPIMESSAGE_RAWTRANSACTION = _LEDGERAPIMESSAGE.nested_types_by_name[
    "RawTransaction"
]
_LEDGERAPIMESSAGE_SIGNEDTRANSACTION = _LEDGERAPIMESSAGE.nested_types_by_name[
    "SignedTransaction"
]
_LEDGERAPIMESSAGE_STATE = _LEDGERAPIMESSAGE.nested_types_by_name["State"]
_LEDGERAPIMESSAGE_TERMS = _LEDGERAPIMESSAGE.nested_types_by_name["Terms"]
_LEDGERAPIMESSAGE_TRANSACTIONDIGEST = _LEDGERAPIMESSAGE.nested_types_by_name[
    "TransactionDigest"
]
_LEDGERAPIMESSAGE_TRANSACTIONRECEIPT = _LEDGERAPIMESSAGE.nested_types_by_name[
    "TransactionReceipt"
]
_LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE = _LEDGERAPIMESSAGE.nested_types_by_name[
    "Get_Balance_Performative"
]
_LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE = (
    _LEDGERAPIMESSAGE.nested_types_by_name["Get_Raw_Transaction_Performative"]
)
_LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE = (
    _LEDGERAPIMESSAGE.nested_types_by_name["Send_Signed_Transaction_Performative"]
)
_LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE = (
    _LEDGERAPIMESSAGE.nested_types_by_name["Get_Transaction_Receipt_Performative"]
)
_LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE = _LEDGERAPIMESSAGE.nested_types_by_name[
    "Balance_Performative"
]
_LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE = _LEDGERAPIMESSAGE.nested_types_by_name[
    "Raw_Transaction_Performative"
]
_LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE = (
    _LEDGERAPIMESSAGE.nested_types_by_name["Transaction_Digest_Performative"]
)
_LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE = (
    _LEDGERAPIMESSAGE.nested_types_by_name["Transaction_Receipt_Performative"]
)
_LEDGERAPIMESSAGE_GET_STATE_PERFORMATIVE = _LEDGERAPIMESSAGE.nested_types_by_name[
    "Get_State_Performative"
]
_LEDGERAPIMESSAGE_STATE_PERFORMATIVE = _LEDGERAPIMESSAGE.nested_types_by_name[
    "State_Performative"
]
_LEDGERAPIMESSAGE_ERROR_PERFORMATIVE = _LEDGERAPIMESSAGE.nested_types_by_name[
    "Error_Performative"
]
LedgerApiMessage = _reflection.GeneratedProtocolMessageType(
    "LedgerApiMessage",
    (_message.Message,),
    {
        "Kwargs": _reflection.GeneratedProtocolMessageType(
            "Kwargs",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_KWARGS,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Kwargs)
            },
        ),
        "RawTransaction": _reflection.GeneratedProtocolMessageType(
            "RawTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_RAWTRANSACTION,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.RawTransaction)
            },
        ),
        "SignedTransaction": _reflection.GeneratedProtocolMessageType(
            "SignedTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_SIGNEDTRANSACTION,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.SignedTransaction)
            },
        ),
        "State": _reflection.GeneratedProtocolMessageType(
            "State",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_STATE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.State)
            },
        ),
        "Terms": _reflection.GeneratedProtocolMessageType(
            "Terms",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TERMS,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Terms)
            },
        ),
        "TransactionDigest": _reflection.GeneratedProtocolMessageType(
            "TransactionDigest",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TRANSACTIONDIGEST,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.TransactionDigest)
            },
        ),
        "TransactionReceipt": _reflection.GeneratedProtocolMessageType(
            "TransactionReceipt",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TRANSACTIONRECEIPT,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.TransactionReceipt)
            },
        ),
        "Get_Balance_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Balance_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Get_Balance_Performative)
            },
        ),
        "Get_Raw_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Raw_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Get_Raw_Transaction_Performative)
            },
        ),
        "Send_Signed_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Send_Signed_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Send_Signed_Transaction_Performative)
            },
        ),
        "Get_Transaction_Receipt_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Transaction_Receipt_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Get_Transaction_Receipt_Performative)
            },
        ),
        "Balance_Performative": _reflection.GeneratedProtocolMessageType(
            "Balance_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Balance_Performative)
            },
        ),
        "Raw_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Raw_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Raw_Transaction_Performative)
            },
        ),
        "Transaction_Digest_Performative": _reflection.GeneratedProtocolMessageType(
            "Transaction_Digest_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Transaction_Digest_Performative)
            },
        ),
        "Transaction_Receipt_Performative": _reflection.GeneratedProtocolMessageType(
            "Transaction_Receipt_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Transaction_Receipt_Performative)
            },
        ),
        "Get_State_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_State_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_GET_STATE_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Get_State_Performative)
            },
        ),
        "State_Performative": _reflection.GeneratedProtocolMessageType(
            "State_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_STATE_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.State_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _LEDGERAPIMESSAGE,
        "__module__": "ledger_api_pb2"
        # @@protoc_insertion_point(class_scope:aea.valory.ledger_api.v1_0_0.LedgerApiMessage)
    },
)
_sym_db.RegisterMessage(LedgerApiMessage)
_sym_db.RegisterMessage(LedgerApiMessage.Kwargs)
_sym_db.RegisterMessage(LedgerApiMessage.RawTransaction)
_sym_db.RegisterMessage(LedgerApiMessage.SignedTransaction)
_sym_db.RegisterMessage(LedgerApiMessage.State)
_sym_db.RegisterMessage(LedgerApiMessage.Terms)
_sym_db.RegisterMessage(LedgerApiMessage.TransactionDigest)
_sym_db.RegisterMessage(LedgerApiMessage.TransactionReceipt)
_sym_db.RegisterMessage(LedgerApiMessage.Get_Balance_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Get_Raw_Transaction_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Send_Signed_Transaction_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Get_Transaction_Receipt_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Balance_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Raw_Transaction_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Transaction_Digest_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Transaction_Receipt_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Get_State_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.State_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Error_Performative)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _LEDGERAPIMESSAGE._serialized_start = 51
    _LEDGERAPIMESSAGE._serialized_end = 2833
    _LEDGERAPIMESSAGE_KWARGS._serialized_start = 1193
    _LEDGERAPIMESSAGE_KWARGS._serialized_end = 1217
    _LEDGERAPIMESSAGE_RAWTRANSACTION._serialized_start = 1219
    _LEDGERAPIMESSAGE_RAWTRANSACTION._serialized_end = 1260
    _LEDGERAPIMESSAGE_SIGNEDTRANSACTION._serialized_start = 1262
    _LEDGERAPIMESSAGE_SIGNEDTRANSACTION._serialized_end = 1309
    _LEDGERAPIMESSAGE_STATE._serialized_start = 1311
    _LEDGERAPIMESSAGE_STATE._serialized_end = 1333
    _LEDGERAPIMESSAGE_TERMS._serialized_start = 1335
    _LEDGERAPIMESSAGE_TERMS._serialized_end = 1357
    _LEDGERAPIMESSAGE_TRANSACTIONDIGEST._serialized_start = 1359
    _LEDGERAPIMESSAGE_TRANSACTIONDIGEST._serialized_end = 1406
    _LEDGERAPIMESSAGE_TRANSACTIONRECEIPT._serialized_start = 1408
    _LEDGERAPIMESSAGE_TRANSACTIONRECEIPT._serialized_end = 1457
    _LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE._serialized_start = 1459
    _LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE._serialized_end = 1521
    _LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE._serialized_start = 1523
    _LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE._serialized_end = 1626
    _LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE._serialized_start = 1629
    _LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE._serialized_end = 1761
    _LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE._serialized_start = 1764
    _LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE._serialized_end = 2004
    _LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE._serialized_start = 2006
    _LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE._serialized_end = 2064
    _LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE._serialized_start = 2066
    _LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE._serialized_end = 2184
    _LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE._serialized_start = 2186
    _LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE._serialized_end = 2313
    _LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE._serialized_start = 2316
    _LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE._serialized_end = 2446
    _LEDGERAPIMESSAGE_GET_STATE_PERFORMATIVE._serialized_start = 2449
    _LEDGERAPIMESSAGE_GET_STATE_PERFORMATIVE._serialized_end = 2595
    _LEDGERAPIMESSAGE_STATE_PERFORMATIVE._serialized_start = 2597
    _LEDGERAPIMESSAGE_STATE_PERFORMATIVE._serialized_end = 2705
    _LEDGERAPIMESSAGE_ERROR_PERFORMATIVE._serialized_start = 2707
    _LEDGERAPIMESSAGE_ERROR_PERFORMATIVE._serialized_end = 2817
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the payloads.py module of the skill."""

from packages.valory.skills.liquidity_rebalancing_abci.payloads import (
    StrategyEvaluationPayload,
    StrategyType,
    TransactionType,
)


def test_strategy_evaluation_payload() -> None:
    """Test `StrategyEvaluationPayload`."""

    strategy = {
        "action": StrategyType.ENTER.value,
        "pair": ["FTM", "BOO"],
        "pool": "0x0000000000000000000000000000",
        "amountETH": 0.1,  # Be careful with floats and determinism here
    }
    payload = StrategyEvaluationPayload(sender="sender", strategy=str(strategy))

    assert payload.sender == "sender"
    assert payload.data == {"strategy": str(strategy)}
    assert payload.transaction_type == TransactionType.STRATEGY_EVALUATION



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""Configurations for APY skill's tests."""

import warnings
from typing import Any, Callable, Dict, List, Tuple, Union
from unittest import mock
from unittest.mock import MagicMock, PropertyMock

import numpy as np
import optuna
import pandas as pd
import pytest
from aea.skills.base import SkillContext
from optuna.distributions import UniformDistribution
from optuna.exceptions import ExperimentalWarning
from optuna.trial import TrialState
from pmdarima import ARIMA
from pmdarima.pipeline import Pipeline
from pmdarima.preprocessing import FourierFeaturizer
from statsmodels.tools.sm_exceptions import ConvergenceWarning

from packages.valory.skills.apy_estimation_abci.ml.forecasting import (
    PoolIdToForecasterType,
    PoolIdToTestReportType,
    PoolIdToTrainDataType,
)
from packages.valory.skills.apy_estimation_abci.ml.optimization import (
    PoolToHyperParamsWithStatusType,
)
from packages.valory.skills.apy_estimation_abci.models import SharedState
from packages.valory.skills.apy_estimation_abci.tools.etl import ResponseItemType


HeaderType = Dict[str, str]
SpecsType = Dict[str, Union[str, int, HeaderType, SkillContext]]


@pytest.fixture
def _common_specs() -> SpecsType:
    return {
        "headers": {"Content-Type": "application/json"},
        "method": "POST",
        "name": "spooky_api",
        "skill_context": SkillContext(),
        "response_key": "data",
        "response_type": "list",
        "retries": 5,
    }


@pytest.fixture
def spooky_specs(_common_specs: SpecsType) -> SpecsType:
    """Spooky specs fixture."""
    return {
        **_common_specs,
        **{
            "api_id": "spookyswap",
            "bundle_id": 1,
            "url": "https://api.thegraph.com/subgraphs/name/eerieeight/spookyswap",
        },
    }


@pytest.fixture
def fantom_specs(_common_specs: SpecsType) -> SpecsType:
    """Fantom specs fixture"""
    return {
        **_common_specs,
        **{
            "api_id": "fantom",
            "url": "https://api.thegraph.com/subgraphs/name/matthewlilley/fantom-blocks",
        },
    }


@pytest.fixture
def eth_price_usd_q() -> str:
    """Query string for fetching ethereum price in USD from SpookySwap."""
    return (
        """
        {
            bundles(
                first: 1,
                block: {number: """
        + str(3830367)
        + """},
                where: {
                    id: """
        + str(1)
        + """
                }
            )
            {ethPrice}
        }
        """
    )


@pytest.fixture
def block_from_timestamp_q() -> str:
    """Query string to get a block from a timestamp."""

    return """
    {
        blocks(
            first: 1,
            orderBy: timestamp,
            orderDirection: asc,
            where: {
                timestamp_gte: 1618735147,
                timestamp_lte: 1618735747
            }
        )
        {
            timestamp
            number
        }
    }
    """


@pytest.fixture
def top_n_pairs_q() -> str:
    """Query to get the first `top_n` pool ids based on their total liquidity."""

    return """
    {
        pairs(
            first: 100,
            orderBy: trackedReserveETH,
            orderDirection: desc
        )
        {id}
    }
    """


@pytest.fixture
def pairs_q() -> str:
    """Query to get data for the first `top_n` pools based on their total liquidity."""

    return (
        """
    {
        pairs(
            where: {id_in:
            [\""""
        + '","'.join(["0xec454eda10accdd66209c57af8c12924556f3abd"])
        + """"]},
            block: {number: """
        + str(3830367)
        + """}
        ) {
            id
            token0 {
                id
                symbol
                name
            }
            token1 {
                id
                symbol
                name
            }
            reserve0
            reserve1
            totalSupply
            reserveETH
            reserveUSD
            trackedReserveETH
            token0Price
            token1Price
            volumeToken0
            volumeToken1
            volumeUSD
            untrackedVolumeUSD
            txCount
            createdAtTimestamp
            createdAtBlockNumber
            liquidityProviderCount
        }
    }
    """
    )


@pytest.fixture
def pool_fields() -> Tuple[str, ...]:
    """The fields of a pool."""
    return (
        "createdAtBlockNumber",
        "createdAtTimestamp",
        "id",
        "liquidityProviderCount",
        "reserve0",
        "reserve1",
        "reserveETH",
        "reserveUSD",
        "token0Price",
        "token1Price",
        "totalSupply",
        "trackedReserveETH",
        "untrackedVolumeUSD",
        "txCount",
        "volumeToken0",
        "volumeToken1",
        "volumeUSD",
        "token0",
        "token1",
    )


@pytest.fixture
def shared_state() -> SharedState:
    """Initialize a test shared state."""
    return SharedState(name="", skill_context=mock.MagicMock())


@pytest.fixture
def no_action() -> Callable[[Any], None]:
    """Create a no-action function."""
    return lambda *_, **__: None


@pytest.fixture
def prepare_batch_task_result() -> pd.DataFrame:
    """Create a dummy prepare batch task result."""
    return pd.DataFrame({"id": ["pool1", "pool2"], "APY": [13.234, 42.34]})


@pytest.fixture
def study() -> MagicMock:
    """Create a dummy optuna Study which has no trials finished."""
    study = MagicMock(trials=(MagicMock(params={"x": 2.0}),))
    type(study).best_params = PropertyMock(
        side_effect=ValueError("Study has no trials finished!")
    )
    return study


@pytest.fixture
def optimize_task_result() -> PoolToHyperParamsWithStatusType:
    """Create a result of the `OptimizeTask`.

    :return: a dummy `Task` Result.
    """
    return {"test1": ({"x": 2.0}, False), "test2": ({"x": 2.4}, True)}


@pytest.fixture
def optimize_task_result_empty() -> optuna.Study:
    """Create an empty result of the `OptimizeTask`.

    :return: a dummy `Task` Result.
    """
    study = optuna.create_study()

    warnings.filterwarnings("ignore", category=ExperimentalWarning)

    trial = optuna.trial.create_trial(
        state=TrialState.WAITING,
        params={"x": 2.0},
        distributions={"x": UniformDistribution(0, 10)},
        value=4.0,
    )

    study.add_trial(trial)

    return study


@pytest.fixture
def optimize_task_result_non_serializable() -> optuna.Study:
    """Create a non-serializable result of the `OptimizeTask`.

    :return: a dummy `Task` Result.
    """
    study = optuna.create_study()

    warnings.filterwarnings("ignore", category=ExperimentalWarning)

    trial = optuna.trial.create_trial(
        params={b"x": 2.0},  # type: ignore
        distributions={b"x": UniformDistribution(0, 10)},  # type: ignore
        value=4.0,
    )

    study.add_trial(trial)

    return study


@pytest.fixture
def observations() -> np.ndarray:
    """Observations of a timeseries."""
    return np.asarray([0] * 20)


@pytest.fixture
def hyperparameters() -> Dict[str, Union[bool, int]]:
    """Hyperparameters for a pipeline."""
    return {
        "p": 1,
        "q": 1,
        "d": 1,
        "m": 3,
        "k": 1,
        "suppress_warnings": True,
    }


@pytest.fixture
def forecaster(hyperparameters: Dict[str, Union[bool, int]]) -> Pipeline:
    """Create a dummy, untrained Pipeline."""
    order = (hyperparameters["p"], hyperparameters["q"], hyperparameters["d"])

    dummy_forecaster = Pipeline(
        [
            ("fourier", FourierFeaturizer(hyperparameters["m"])),
            (
                "arima",
                ARIMA(order),
            ),
        ]
    )

    return dummy_forecaster


@pytest.fixture
def trained_forecaster(forecaster: Pipeline, observations: np.ndarray) -> Pipeline:
    """Create a dummy, trained Pipeline."""
    warnings.filterwarnings("ignore", category=ConvergenceWarning)
    forecaster.fit(observations)

    return forecaster


@pytest.fixture
def train_task_input(observations: np.ndarray) -> PoolIdToTrainDataType:
    """Create an input of the `TrainTask`."""
    return {f"pool{i}.csv": observations for i in range(3)}


@pytest.fixture
def train_task_result(trained_forecaster: Pipeline) -> PoolIdToForecasterType:
    """Create a result of the `TrainTask`.

    :param trained_forecaster: a trained, dummy forecaster.
    :return: a dummy `TrainTask` Result.
    """
    return {f"pool{i}.joblib": trained_forecaster for i in range(3)}


@pytest.fixture
def _test_task_result() -> PoolIdToTestReportType:
    """Create a result of the `TestTask`.

    :return: a dummy `Task` Result.
    """
    return {f"pool{i}": {"test": "test"} for i in range(3)}


@pytest.fixture
def _test_task_result_non_serializable() -> bytes:
    """Create a non-serializable result of the `TestTask`.

    :return: a dummy `Task` Result.
    """
    return b"non-serializable"


@pytest.fixture
def historical_data() -> Dict[str, List[Union[None, Dict[str, str], int, str, float]]]:
    """Create dummy transformed historical data"""
    return {
        "createdAtBlockNumber": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "createdAtTimestamp": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "id": [
            "0x2b4c76d0dc16be1c31d4c1dc53bf9b45987fc75c",
            "x2",
            "x3",
            "x4",
            "x5",
            "0x2b4c76d0dc16be1c31d4c1dc53bf9b45987fc75c",
            "0x2b4c76d0dc16be1c31d4c1dc53bf9b45987fc75c",
            "0x2b4c76d0dc16be1c31d4c1dc53bf9b45987fc75c",
            "0x2b4c76d0dc16be1c31d4c1dc53bf9b45987fc75c",
            "0x2b4c76d0dc16be1c31d4c1dc53bf9b45987fc75c",
        ],
        "liquidityProviderCount": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "reserve0": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "reserve1": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "reserveETH": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "reserveUSD": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "token0Price": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "token1Price": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "totalSupply": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "trackedReserveETH": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "untrackedVolumeUSD": [1.4, 2.4, 3.4, 4.4, 5.4, 6.4, 7.4, 8.0, 9.4, 10.7],
        "txCount": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "volumeToken0": [1.2, 2.4, 3.0, 4.6, 5.8, 6.3, 7.5, 8.2, 9.1, 10.7],
        "volumeToken1": [1.2, 2.4, 3.0, 4.6, 5.8, 6.3, 7.5, 8.2, 9.1, 10.7],
        "volumeUSD": [1.2, 2.4, 3.0, 4.6, 5.8, 6.3, 7.5, 8.2, 9.1, 10.7],
        "forTimestamp": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "blockNumber": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "blockTimestamp": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "ethPrice": [1.2, 2.3, 3.4, 4.45634, 5.2, 6.0, 7.246, 8.26, 9.123, 10.56],
        "token0": [
            {"id": "x", "name": "x", "symbol": "x"},
            {"id": "k", "name": "k", "symbol": "k"},
            {"id": "l", "name": "l", "symbol": "l"},
            {"id": "t", "name": "t", "symbol": "t"},
            {"id": "v", "name": "v", "symbol": "v"},
            {"id": "x", "name": "x", "symbol": "x"},
            {"id": "x", "name": "x", "symbol": "x"},
            {"id": "x", "name": "x", "symbol": "x"},
            {"id": "x", "name": "x", "symbol": "x"},
            {"id": "x", "name": "x", "symbol": "x"},
        ],
        "token1": [
            {"id": "y", "name": "y", "symbol": "y"},
            {"id": "m", "name": "m", "symbol": "m"},
            {"id": "r", "name": "r", "symbol": "r"},
            {"id": "y", "name": "y", "symbol": "y"},
            {"id": "b", "name": "b", "symbol": "b"},
            {"id": "y", "name": "y", "symbol": "y"},
            {"id": "y", "name": "y", "symbol": "y"},
            {"id": "y", "name": "y", "symbol": "y"},
            {"id": "y", "name": "y", "symbol": "y"},
            {"id": "y", "name": "y", "symbol": "y"},
        ],
    }


@pytest.fixture
def transformed_historical_data_no_datetime_conversion(
    historical_data: Dict[str, List[Union[None, Dict[str, str], int, str, float]]]
) -> pd.DataFrame:
    """Create dummy transformed historical data"""
    data_copy = historical_data.copy()
    data_copy.update(
        {
            "APY": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.1],
            "currentChange": [None, None, None, None, None, 5.1, 1.2, 0.7, 0.9, 1.6],
            "token0ID": ["x", "k", "l", "t", "v", "x", "x", "x", "x", "x"],
            "token0Name": ["x", "k", "l", "t", "v", "x", "x", "x", "x", "x"],
            "token0Symbol": ["x", "k", "l", "t", "v", "x", "x", "x", "x", "x"],
            "token1ID": ["y", "m", "r", "y", "b", "y", "y", "y", "y", "y"],
            "token1Name": ["y", "m", "r", "y", "b", "y", "y", "y", "y", "y"],
            "token1Symbol": ["y", "m", "r", "y", "b", "y", "y", "y", "y", "y"],
            "pairName": [
                "x - y",
                "k - m",
                "l - r",
                "t - y",
                "v - b",
                "x - y",
                "x - y",
                "x - y",
                "x - y",
                "x - y",
            ],
            "updatedVolumeUSD": [1.2, 2.4, 3.0, 4.6, 5.8, 6.3, 7.5, 8.2, 9.1, 10.7],
            "updatedReserveUSD": [
                1.68,
                5.52,
                11.559999999999999,
                19.607896,
                28.080000000000002,
                38.400000000000006,
                53.620400000000004,
                66.08,
                85.75619999999999,
                112.992,
            ],
        }
    )

    for i in range(2):
        del data_copy[f"token{i}"]

    return pd.DataFrame(data_copy)


@pytest.fixture
def transformed_historical_data(
    transformed_historical_data_no_datetime_conversion: pd.DataFrame,
) -> pd.DataFrame:
    """Create dummy transformed historical data"""
    transformed_historical_data_no_datetime_conversion[
        "blockTimestamp"
    ] = pd.to_datetime(
        transformed_historical_data_no_datetime_conversion["blockTimestamp"], unit="s"
    )

    return transformed_historical_data_no_datetime_conversion


@pytest.fixture
def batch() -> ResponseItemType:
    """Create a dummy batch of data."""
    pool1_batch: Dict[str, Union[str, Dict[str, str]]] = {
        "createdAtBlockNumber": "1",
        "createdAtTimestamp": "1",
        "id": "0x2b4c76d0dc16be1c31d4c1dc53bf9b45987fc75c",
        "liquidityProviderCount": "1",
        "reserve0": "1.4",
        "reserve1": "1.4",
        "reserveETH": "1.4",
        "reserveUSD": "1.4",
        "token0Price": "1.4",
        "token1Price": "1.4",
        "totalSupply": "1.4",
        "trackedReserveETH": "1.4",
        "untrackedVolumeUSD": "1.4",
        "txCount": "1",
        "volumeToken0": "1.2",
        "volumeToken1": "1.2",
        "volumeUSD": "1.2",
        "forTimestamp": "1",
        "blockNumber": "1",
        "blockTimestamp": "100000000",
        "ethPrice": "1.2",
        "token0ID": "x",
        "token0Name": "x",
        "token0Symbol": "x",
        "token1ID": "y",
        "token1Name": "y",
        "token1Symbol": "y",
        "pairName": "x - y",
        "updatedVolumeUSD": "1.2",
        "updatedReserveUSD": "1.68",
        "APY": "0.1",
        "token0": {"id": "x", "name": "x", "symbol": "x"},
        "token1": {"id": "y", "name": "y", "symbol": "y"},
    }
    pool2_batch = pool1_batch.copy()
    pool2_batch["id"] = "x3"
    return [pool1_batch, pool2_batch]


class DummyPipeline(Pipeline):
    """A dummy pipeline."""

    def __init__(self) -> None:
        """Initialize Dummy Pipeline."""
        super().__init__([])
        self.updated: bool = False

    def _validate_steps(self) -> None:
        """Dummy steps validation."""

    @staticmethod
    def predict(*args: Any) -> np.ndarray:
        """Predict `steps_forward` timesteps in the future.

        :param args: the args accepted by `pmdarima.Pipeline.predict`.
        :return: a `numpy` array with the dummy predictions.
        """
        return np.ones(args[0])

    def update(self, *args: Any) -> None:
        """Update the dummy pipeline."""
        if len(args[0]):
            self.updated = True


def is_list_of_strings(lst: Any) -> bool:
    """Check if arg is a list of strings."""
    res = False

    if lst and isinstance(lst, list):
        res = all(isinstance(elem, str) for elem in lst)

    return res



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/price_estimation_abci skill's behaviours."""
import copy
import json
import sys
import time
from pathlib import Path
from typing import Any, Dict, Union, cast
from unittest import mock

import atheris  # type: ignore
import pytest
from aea.exceptions import AEAEnforceError
from aea.helpers.transaction.base import RawTransaction, SignedMessage

from packages.open_aea.protocols.signing import SigningMessage
from packages.valory.contracts.gnosis_safe.contract import (
    PUBLIC_ID as GNOSIS_SAFE_CONTRACT_ID,
)
from packages.valory.contracts.offchain_aggregator.contract import (
    PUBLIC_ID as ORACLE_CONTRACT_ID,
)
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    make_degenerate_behaviour,
)
from packages.valory.skills.price_estimation_abci.behaviours import (
    EstimateBehaviour,
    ObserveBehaviour,
    TransactionHashBehaviour,
    pack_for_server,
)
from packages.valory.skills.price_estimation_abci.payloads import ObservationPayload
from packages.valory.skills.price_estimation_abci.rounds import (
    Event,
    FinishedPriceAggregationRound,
)
from packages.valory.skills.price_estimation_abci.rounds import (
    SynchronizedData as PriceEstimationSynchronizedSata,
)

from tests.conftest import ROOT_DIR
from tests.test_skills.base import FSMBehaviourBaseCase


DRAND_VALUE = {
    "round": 1416669,
    "randomness": "f6be4bf1fa229f22340c1a5b258f809ac4af558200775a67dacb05f0cb258a11",
    "signature": (
        "b44d00516f46da3a503f9559a634869b6dc2e5d839e46ec61a090e3032172954929a5"
        "d9bd7197d7739fe55db770543c71182562bd0ad20922eb4fe6b8a1062ed21df3b68de"
        "44694eb4f20b35262fa9d63aa80ad3f6172dd4d33a663f21179604"
    ),
    "previous_signature": (
        "903c60a4b937a804001032499a855025573040cb86017c38e2b1c3725286756ce8f33"
        "61188789c17336beaf3f9dbf84b0ad3c86add187987a9a0685bc5a303e37b008fba8c"
        "44f02a416480dd117a3ff8b8075b1b7362c58af195573623187463"
    ),
}


class DummyRoundId:
    """Dummy class for setting round_id for exit condition."""

    round_id: str

    def __init__(self, round_id: str) -> None:
        """Dummy class for setting round_id for exit condition."""
        self.round_id = round_id


class PriceEstimationFSMBehaviourBaseCase(FSMBehaviourBaseCase):
    """Base case for testing PriceEstimation FSMBehaviour."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "valory", "skills", "price_estimation_abci"
    )


class TestObserveBehaviour(PriceEstimationFSMBehaviourBaseCase):
    """Test ObserveBehaviour."""

    def test_observer_behaviour(
        self,
    ) -> None:
        """Run tests."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            ObserveBehaviour.behaviour_id,
            PriceEstimationSynchronizedSata(
                AbciAppDB(initial_data=dict(estimate=[1.0])),
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == ObserveBehaviour.behaviour_id
        )
        self.behaviour.act_wrapper()
        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                url="https://api.coinbase.com/v2/prices/BTC-USD/buy",
                headers="",
                version="",
                body=b"",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=json.dumps({"data": {"amount": 54566}}).encode("utf-8"),
            ),
        )
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(Event.DONE)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == EstimateBehaviour.behaviour_id

    def test_observer_behaviour_retries_exceeded(
        self,
    ) -> None:
        """Run tests."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            ObserveBehaviour.behaviour_id,
            PriceEstimationSynchronizedSata(
                AbciAppDB(initial_data=dict(estimate=[1.0])),
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == ObserveBehaviour.behaviour_id
        )
        with mock.patch.object(
            self.behaviour.context.price_api,
            "is_retries_exceeded",
            return_value=True,
        ):
            self.behaviour.act_wrapper()
            behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
            assert behaviour.behaviour_id == ObserveBehaviour.behaviour_id
            self._test_done_flag_set()

    def test_observed_value_none(
        self,
    ) -> None:
        """Test when `observed` value is none."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            ObserveBehaviour.behaviour_id,
            PriceEstimationSynchronizedSata(
                AbciAppDB(initial_data=dict()),
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == ObserveBehaviour.behaviour_id
        )
        self.behaviour.act_wrapper()
        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                url="https://api.coinbase.com/v2/prices/BTC-USD/buy",
                headers="",
                version="",
                body=b"",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=b"",
            ),
        )
        time.sleep(1)
        self.behaviour.act_wrapper()

    def test_clean_up(
        self,
    ) -> None:
        """Test when `observed` value is none."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            ObserveBehaviour.behaviour_id,
            PriceEstimationSynchronizedSata(
                AbciAppDB(initial_data=dict()),
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == ObserveBehaviour.behaviour_id
        )
        self.behaviour.context.price_api._retries_attempted = 1
        assert self.behaviour.current_behaviour is not None
        self.behaviour.current_behaviour.clean_up()
        assert self.behaviour.context.price_api._retries_attempted == 0


class TestEstimateBehaviour(PriceEstimationFSMBehaviourBaseCase):
    """Test EstimateBehaviour."""

    def test_estimate(
        self,
    ) -> None:
        """Test estimate behaviour."""

        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=EstimateBehaviour.behaviour_id,
            synchronized_data=PriceEstimationSynchronizedSata(
                AbciAppDB(
                    initial_data=dict(
                        participant_to_observations=[
                            {"a": ObservationPayload(sender="a", observation=1.0)}
                        ]
                    ),
                ),
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == EstimateBehaviour.behaviour_id
        )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(Event.DONE)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == TransactionHashBehaviour.behaviour_id


def mock_to_server_message_flow(
    self: "TestTransactionHashBehaviour",
    this_period_count: int,
    prev_tx_hash: str,
) -> None:
    """Mock to server message flow"""

    self.behaviour.context.logger.info("Mocking to server message flow")
    # note that although this is a dict, order matters for the test
    data = {
        "period_count": this_period_count,
        "agent_address": "test_agent_address",
        "estimate": 1.0,
        "prev_tx_hash": prev_tx_hash,
        "observations": {"agent1": 0.0},
        "data_source": "coinbase",
        "unit": "BTC:USD",
    }
    behaviour = self.behaviour.current_behaviour  # type: ignore
    participants = behaviour.synchronized_data.sorted_participants  # type: ignore
    decimals = behaviour.params.oracle_params["decimals"]  # type: ignore
    data["package"] = pack_for_server(participants, decimals, **data).hex()  # type: ignore
    data["signature"] = "stub_signature"

    request_kwargs: Dict[str, Union[str, bytes]] = dict(
        method="POST",
        url="http://192.168.2.17:9999/deposit",
        headers="",
        version="",
        body=str(data).encode("utf-8"),
    )

    response_kwargs = dict(
        version="",
        status_code=201,
        status_text="",
        headers="",
        body=b"",
    )

    # mock signature acquisition
    self.behaviour.act_wrapper()
    self.mock_signing_request(
        request_kwargs=dict(
            performative=SigningMessage.Performative.SIGN_MESSAGE,
        ),
        response_kwargs=dict(
            performative=SigningMessage.Performative.SIGNED_MESSAGE,
            signed_message=SignedMessage(ledger_id="ethereum", body="stub_signature"),
        ),
    )

    self.behaviour.act_wrapper()
    self.mock_http_request(request_kwargs, response_kwargs)
    self.behaviour.act_wrapper()


def get_valid_server_data() -> Dict[str, Any]:
    """Get valid server data"""

    participants = [
        "0x3C44CdDdB6a900fa2b585dd299e03d12FA4293BC",
        "0x70997970C51812dc3A010C7d01b50e0d17dc79C8",
        "0x90F79bf6EB2c4f870365E785982E1f101E93b906",
        "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
    ]
    period_count = 123456789
    estimate = 43974.960019901744
    observations = {
        "0x3C44CdDdB6a900fa2b585dd299e03d12FA4293BC": 44251.11,
        "0x70997970C51812dc3A010C7d01b50e0d17dc79C8": 43964.3900398035,
        "0x90F79bf6EB2c4f870365E785982E1f101E93b906": 43985.53,
        "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266": 43949.0,
    }
    data_source = "coinbase"
    unit = "BTC:USD"
    return dict(
        participants=participants,
        period_count=period_count,
        estimate=estimate,
        observations=observations,
        data_source=data_source,
        unit=unit,
    )


@pytest.mark.parametrize(
    "broadcast_to_server, this_period_count", ((True, 0), (False, 0), (True, 1))
)
class TestTransactionHashBehaviour(PriceEstimationFSMBehaviourBaseCase):
    """Test TransactionHashBehaviour."""

    def test_estimate(
        self,
        broadcast_to_server: bool,
        this_period_count: int,
    ) -> None:
        """Test estimate behaviour."""

        # change setting, mock message flow with and without broadcast to server
        behaviour_params = self.behaviour.current_behaviour.params  # type: ignore
        behaviour_params.is_broadcasting_to_server = broadcast_to_server  # type: ignore

        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=TransactionHashBehaviour.behaviour_id,
            synchronized_data=PriceEstimationSynchronizedSata(
                AbciAppDB(
                    initial_data=dict(
                        most_voted_estimate=[1.0],
                        safe_contract_address=["safe_contract_address"],
                        oracle_contract_address=[
                            "0x77E9b2EF921253A171Fa0CB9ba80558648Ff7215"
                        ],
                    ),
                )
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == TransactionHashBehaviour.behaviour_id
        )
        self.behaviour.act_wrapper()
        self.mock_contract_api_request(
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ),
            contract_id=str(ORACLE_CONTRACT_ID),
            response_kwargs=dict(
                performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                callable="get_latest_transmission_details",
                raw_transaction=RawTransaction(
                    ledger_id="ethereum", body={"epoch_": 1, "round_": 1}
                ),
            ),
        )
        self.mock_contract_api_request(
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ),
            contract_id=str(ORACLE_CONTRACT_ID),
            response_kwargs=dict(
                performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                callable="get_transmit_data",
                raw_transaction=RawTransaction(
                    ledger_id="ethereum", body={"data": b"data"}  # type: ignore
                ),
            ),
        )

        tx_hashes = ["", "0x_prev_cycle_tx_hash"]
        synchronized_data = self.behaviour.current_behaviour.synchronized_data  # type: ignore
        period_data = synchronized_data.db.get_latest()
        period_data.update(
            {
                "participants": {"agent1"},
                "participant_to_observations": {"agent1": 1.0},
                "most_voted_estimate": 1.0,
                "final_tx_hash": tx_hashes[1],
            }
        )

        # add new cycle, and dummy period data
        next_period_data = copy.deepcopy(period_data)
        next_period_data["final_tx_hash"] = tx_hashes[0]

        synchronized_data.db.update(
            **period_data,
        )

        if this_period_count != 0:
            synchronized_data.db.create(
                **AbciAppDB.data_to_lists(next_period_data),
            )

        self.mock_contract_api_request(
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ),
            contract_id=str(GNOSIS_SAFE_CONTRACT_ID),
            response_kwargs=dict(
                performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                callable="get_deploy_transaction",
                raw_transaction=RawTransaction(
                    ledger_id="ethereum",
                    body={
                        "tx_hash": "0xb0e6add595e00477cf347d09797b156719dc5233283ac76e4efce2a674fe72d9"
                    },
                ),
            ),
        )

        if broadcast_to_server:
            tx_hash = tx_hashes[this_period_count]
            mock_to_server_message_flow(self, this_period_count, tx_hash)

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(Event.DONE)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert (
            behaviour.behaviour_id
            == make_degenerate_behaviour(
                FinishedPriceAggregationRound.round_id
            ).behaviour_id
        )


class TestPackForServer(PriceEstimationFSMBehaviourBaseCase):
    """Test packaging of data for signing by agents"""

    @pytest.mark.parametrize(
        "mutation, remains_valid",
        (
            ({}, True),  # do nothing
            ({"participants": ("",) * 4}, True),
            ({"period_count": (1 << 256) - 1}, True),
            ({"period_count": 1 << 256}, False),
            ({"estimate": f"{'9'*30}.{'9'*1}"}, True),
            ({"estimate": f"{'9'*30}.{'9'*2}"}, False),
            ({"data_source": "a" * 32}, True),
            ({"data_source": "a" * 33}, False),
            ({"unit": "a" * 32}, True),
            ({"unit": "a" * 33}, False),
        ),
    )
    def test_pack_for_server(
        self,
        mutation: Dict[str, Any],
        remains_valid: bool,
    ) -> None:
        """Test packaging valid and invalid data"""

        decimals = self.behaviour.current_behaviour.params.oracle_params["decimals"]  # type: ignore
        kwargs = get_valid_server_data()
        kwargs.update({"decimals": decimals})
        kwargs.update(mutation)

        if remains_valid:
            package = pack_for_server(**kwargs)
            assert isinstance(package, bytes) and len(package) == 256
        else:
            with pytest.raises((OverflowError, AEAEnforceError)):
                pack_for_server(**kwargs)


@pytest.mark.skip
def test_fuzz_pack_for_server() -> None:
    """Test fuzz pack_for_server."""

    @atheris.instrument_func
    def fuzz_pack_for_server(input_bytes: bytes) -> None:
        """Fuzz pack_for_server."""
        fdp = atheris.FuzzedDataProvider(input_bytes)
        participants = [fdp.ConsumeStr(12) for _ in range(4)]
        decimals = fdp.ConsumeInt(4)
        period_count = fdp.ConsumeInt(4)
        estimate = fdp.ConsumeFloat()
        observations = {p: fdp.ConsumeFloat() for p in participants}
        data_source = fdp.ConsumeStr(12)
        unit = fdp.ConsumeStr(12)
        pack_for_server(
            participants,
            decimals,
            period_count,
            estimate,
            observations,
            data_source,
            unit,
        )

    atheris.instrument_all()
    atheris.Setup(sys.argv, fuzz_pack_for_server)
    atheris.Fuzz()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the shared state for the registration abci skill."""

from typing import Any

from packages.valory.skills.abstract_round_abci.models import BaseParams
from packages.valory.skills.abstract_round_abci.models import (
    BenchmarkTool as BaseBenchmarkTool,
)
from packages.valory.skills.abstract_round_abci.models import Requests as BaseRequests
from packages.valory.skills.abstract_round_abci.models import (
    SharedState as BaseSharedState,
)
from packages.valory.skills.registration_abci.rounds import AgentRegistrationAbciApp


class SharedState(BaseSharedState):
    """Keep the current shared state of the skill."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the state."""
        super().__init__(*args, abci_app_cls=AgentRegistrationAbciApp, **kwargs)


Params = BaseParams
Requests = BaseRequests
BenchmarkTool = BaseBenchmarkTool



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Decode AEA messages from Tendermint protobuf messages."""

# isort: skip_file  # noqa

from typing import Callable, Optional, Tuple, cast

from aea.exceptions import enforce
from google.protobuf.timestamp_pb2 import Timestamp as TimestampPb

from packages.valory.connections.abci.dialogues import AbciDialogue, AbciDialogues
from packages.valory.connections.abci.tendermint.abci.types_pb2 import (  # type: ignore
    Evidence as EvidencePb,
)
from packages.valory.connections.abci.tendermint.abci.types_pb2 import (
    LastCommitInfo as LastCommitInfoPb,
)
from packages.valory.connections.abci.tendermint.abci.types_pb2 import Request
from packages.valory.connections.abci.tendermint.abci.types_pb2 import (
    Validator as ValidatorPb,
)
from packages.valory.connections.abci.tendermint.types.types_pb2 import (  # type: ignore
    Header as HeaderPb,
)
from packages.valory.protocols.abci import AbciMessage
from packages.valory.protocols.abci.custom_types import (
    BlockID,
    CheckTxType,
    CheckTxTypeEnum,
    ConsensusParams,
    Evidence,
    Evidences,
    Header,
    LastCommitInfo,
    PartSetHeader,
    Snapshot,
    Timestamp,
    Validator,
    ValidatorUpdates,
    ValidatorUpdate,
)


from packages.valory.connections.abci.tendermint.abci.types_pb2 import (  # isort:skip
    ConsensusParams as ConsensusParamsPb,
)
from packages.valory.connections.abci.tendermint.abci.types_pb2 import (  # isort:skip
    ValidatorUpdate as ValidatorUpdatePb,
)


from packages.valory.connections.abci.tendermint.types.types_pb2 import (  # type: ignore
    BlockID as BlockIDPb,
)


class _TendermintProtocolDecoder:
    """
    Decoder called by the server to process requests from the TCP connection with Tendermint.

    It translates from Tendermint's ABCI Protobuf messages into the AEA's ABCI protocol messages.
    """

    @classmethod
    def process(
        cls, message: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Optional[Tuple[AbciMessage, AbciDialogue]]:
        """Process an ABCI request or response."""
        is_request = isinstance(message, Request)
        enforce(is_request, "only Request messages are allowed")
        message_type = f"request_{message.WhichOneof('value')}"
        handler: Callable[
            [Request, AbciDialogues, str], Tuple[AbciMessage, AbciDialogue]
        ] = getattr(cls, message_type, cls.no_match)
        result = handler(message, dialogues, counterparty)
        return result

    @classmethod
    def request_echo(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode an echo request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        echo = request.echo
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_ECHO,
            counterparty=counterparty,
            message=echo.message,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_flush(
        cls, _request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a flush request.

        :param _request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_FLUSH,
            counterparty=counterparty,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_info(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a info request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        info = request.info
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_INFO,
            counterparty=counterparty,
            version=info.version,
            block_version=info.block_version,
            p2p_version=info.p2p_version,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_set_option(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a set_option request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        set_option = request.set_option
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_SET_OPTION,
            counterparty=counterparty,
            option_key=set_option.key,
            option_value=set_option.value,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_init_chain(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a init_chain request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        init_chain = request.init_chain
        timestamp = cls._decode_timestamp(init_chain.time)
        chain_id = init_chain.chain_id
        consensus_params = (
            cls._decode_consensus_params(init_chain.consensus_params)
            if init_chain.consensus_params is not None
            else None
        )
        validators = ValidatorUpdates(
            [
                cls._decode_validator_update(validator_update_pb)
                for validator_update_pb in list(init_chain.validators)
            ]
        )
        app_state_bytes = init_chain.app_state_bytes
        initial_height = init_chain.initial_height

        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_INIT_CHAIN,
            counterparty=counterparty,
            time=timestamp,
            chain_id=chain_id,
            validators=validators,
            app_state_bytes=app_state_bytes,
            initial_height=initial_height,
        )
        if consensus_params is not None:
            abci_message.set("consensus_params", consensus_params)
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_begin_block(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a begin_block request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        begin_block = request.begin_block
        hash_ = begin_block.hash
        header = cls._decode_header(begin_block.header)
        last_commit_info = cls._decode_last_commit_info(begin_block.last_commit_info)
        evidences = [
            cls._decode_evidence(byzantine_validator)
            for byzantine_validator in list(begin_block.byzantine_validators)
        ]
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_BEGIN_BLOCK,
            counterparty=counterparty,
            hash=hash_,
            header=header,
            last_commit_info=last_commit_info,
            byzantine_validators=Evidences(evidences),
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_check_tx(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a check_tx request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        check_tx = request.check_tx
        transaction = check_tx.tx
        check_tx_type = CheckTxType(CheckTxTypeEnum(check_tx.type))
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_CHECK_TX,
            counterparty=counterparty,
            tx=transaction,
            type=check_tx_type,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_deliver_tx(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a deliver_tx request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_DELIVER_TX,
            counterparty=counterparty,
            tx=request.deliver_tx.tx,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_query(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a query request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_QUERY,
            counterparty=counterparty,
            query_data=request.query.data,
            path=request.query.path,
            height=request.query.height,
            prove=request.query.prove,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_commit(
        cls, _request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode a commit request.

        :param _request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_COMMIT,
            counterparty=counterparty,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_end_block(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """
        Decode an end_block request.

        :param request: the request.
        :param dialogues: the dialogues object.
        :param counterparty: the counterparty.
        :return: the AbciMessage request.
        """
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_END_BLOCK,
            counterparty=counterparty,
            height=request.end_block.height,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_list_snapshots(
        cls, _request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """Decode a list_snapshots request."""
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_LIST_SNAPSHOTS,
            counterparty=counterparty,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_offer_snapshot(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """Decode a offer_snapshot request."""
        offer_snapshot = request.offer_snapshot
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_OFFER_SNAPSHOT,
            counterparty=counterparty,
            snapshot=Snapshot(
                offer_snapshot.snapshot.height,
                offer_snapshot.snapshot.format,
                offer_snapshot.snapshot.chunks,
                offer_snapshot.snapshot.hash,
                offer_snapshot.snapshot.metadata,
            ),
            app_hash=offer_snapshot.app_hash,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_load_snapshot_chunk(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """Decode a load_snapshot_chunk request."""
        load_snapshot_chunk = request.load_snapshot_chunk
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_LOAD_SNAPSHOT_CHUNK,
            counterparty=counterparty,
            height=load_snapshot_chunk.height,
            format=load_snapshot_chunk.format,
            chunk_index=load_snapshot_chunk.chunk,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def request_apply_snapshot_chunk(
        cls, request: Request, dialogues: AbciDialogues, counterparty: str
    ) -> Tuple[AbciMessage, AbciDialogue]:
        """Decode a apply_snapshot_chunk request."""
        apply_snapshot_chunk = request.apply_snapshot_chunk
        abci_message, abci_dialogue = dialogues.create(
            performative=AbciMessage.Performative.REQUEST_APPLY_SNAPSHOT_CHUNK,
            counterparty=counterparty,
            index=apply_snapshot_chunk.index,
            chunk=apply_snapshot_chunk.chunk,
            chunk_sender=apply_snapshot_chunk.sender,
        )
        return cast(AbciMessage, abci_message), cast(AbciDialogue, abci_dialogue)

    @classmethod
    def no_match(
        cls, _request: Request, _dialogues: AbciDialogues, _counterparty: str
    ) -> None:  # pragma: nocover
        """Handle the case in which the request is not supported."""
        return None

    @classmethod
    def _decode_timestamp(cls, timestamp_tendermint_pb: TimestampPb) -> Timestamp:
        """Decode a timestamp object."""
        return Timestamp(
            timestamp_tendermint_pb.seconds,
            timestamp_tendermint_pb.nanos,
        )

    @classmethod
    def _decode_consensus_params(
        cls, consensus_params_tendermint_pb: ConsensusParamsPb
    ) -> ConsensusParams:
        """Decode a ConsensusParams object."""
        return ConsensusParams.decode(consensus_params_tendermint_pb)

    @classmethod
    def _decode_validator_update(
        cls, validator_update_pb: ValidatorUpdatePb
    ) -> ValidatorUpdate:
        """Decode a ValidatorUpdate object."""
        return ValidatorUpdate.decode(validator_update_pb)

    @classmethod
    def _decode_header(cls, header_tendermint_pb: HeaderPb) -> Header:
        """Decode a Header object."""
        return Header.decode(header_tendermint_pb)

    @classmethod
    def _decode_block_id(cls, block_id_pb: BlockIDPb) -> BlockID:  # pragma: nocover
        """Decode a Block ID object."""
        part_set_header_pb = block_id_pb.part_set_header
        part_set_header = PartSetHeader(
            part_set_header_pb.index, part_set_header_pb.bytes
        )
        return BlockID(block_id_pb.hash, part_set_header)

    @classmethod
    def _decode_last_commit_info(
        cls, last_commit_info_tendermint_pb: LastCommitInfoPb
    ) -> LastCommitInfo:
        """Decode a LastCommitInfo object."""
        return LastCommitInfo.decode(last_commit_info_tendermint_pb)

    @classmethod
    def _decode_validator(
        cls, validator_pb: ValidatorPb
    ) -> Validator:  # pragma: nocover
        """Decode a Validator object."""
        return Validator(validator_pb.address, validator_pb.power)

    @classmethod
    def _decode_evidence(cls, evidence_pb: EvidencePb) -> Evidence:  # pragma: nocover
        """Decode an Evidence object."""
        return Evidence.decode(evidence_pb)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the class to connect to a ERC20 contract."""
import logging
from typing import Any, List, Optional, cast

from aea.common import JSONLike
from aea.configurations.base import PublicId
from aea.contracts.base import Contract
from aea_ledger_ethereum import EthereumApi


PUBLIC_ID = PublicId.from_str("valory/uniswap_v2_erc20:0.1.0")

_logger = logging.getLogger(
    f"aea.packages.{PUBLIC_ID.author}.contracts.{PUBLIC_ID.name}.contract"
)


def snake_to_camel(string: str) -> str:
    """Convert snake_case to camelCase"""

    if "_" in string:
        camel_case = string.split("_")
        for i in range(1, len(camel_case)):
            camel_case[i] = camel_case[i][0].upper() + camel_case[i][1:]
        string = ("").join(camel_case)
    return string


# pylint: disable=too-many-arguments,invalid-name
class UniswapV2ERC20Contract(Contract):
    """The Uniswap V2 ERC-20 contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_method_data(
        cls,
        ledger_api: EthereumApi,
        contract_address: str,
        method_name: str,
        **kwargs: Any,
    ) -> JSONLike:
        """
        Get a contract call encoded data.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param method_name: the contract method name
        :param kwargs: the contract method args
        :return: the tx  # noqa: DAR202
        """
        instance = cls.get_instance(ledger_api, contract_address)

        # Ensure the method name and its arguments are camel-cased
        # to match the contract interface
        method_name = snake_to_camel(method_name)
        kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}

        # Get an ordered argument list from the method's abi
        method = instance.get_function_by_name(method_name)
        input_names = [i["name"] for i in method.abi["inputs"]]
        args = [kwargs[i] for i in input_names]

        # Encode and return the contract call
        data = instance.encodeABI(fn_name=method_name, args=args)
        return {"data": bytes.fromhex(data[2:])}  # type: ignore

    @classmethod
    def approve(
        cls,
        ledger_api: EthereumApi,
        contract_address: str,
        spender_address: str,
        value: int,
        **kwargs: Any,
    ) -> Optional[JSONLike]:
        """Set the allowance for spender_address on behalf of sender_address."""
        contract_instance = cls.get_instance(ledger_api, contract_address)

        return ledger_api.build_transaction(
            contract_instance,
            "approve",
            method_args=dict(
                spender=spender_address,
                value=value,
            ),
            tx_args=kwargs,
        )

    @classmethod
    def transfer(
        cls,
        ledger_api: EthereumApi,
        contract_address: str,
        to_address: str,
        value: int,
        **tx_args: Any,
    ) -> Optional[JSONLike]:
        """Transfer funds from sender_address to to_address."""
        contract_instance = cls.get_instance(ledger_api, contract_address)

        return ledger_api.build_transaction(
            contract_instance,
            "transfer",
            method_args={
                "to": to_address,
                "value": value,
            },
            tx_args=tx_args,
        )

    @classmethod
    def transfer_from(
        cls,
        ledger_api: EthereumApi,
        contract_address: str,
        from_address: str,
        to_address: str,
        value: int,
        **tx_args: Any,
    ) -> Optional[JSONLike]:
        """As sender_address (third-party) transfer funds from from_address to to_address."""
        contract_instance = cls.get_instance(ledger_api, contract_address)

        return ledger_api.build_transaction(
            contract_instance,
            "transferFrom",
            method_args={
                "from": from_address,
                "to": to_address,
                "value": value,
            },
            tx_args=tx_args,
        )

    @classmethod
    def permit(
        cls,
        ledger_api: EthereumApi,
        contract_address: str,
        owner_address: str,
        spender_address: str,
        value: int,
        deadline: int,
        v: int,
        r: bytes,
        s: bytes,
        **kwargs: Any,
    ) -> Optional[JSONLike]:
        """Sets the allowance for a spender on behalf of owner where approval is granted via a signature. Sender can differ from owner."""
        contract_instance = cls.get_instance(ledger_api, contract_address)

        return ledger_api.build_transaction(
            contract_instance,
            "permit",
            method_args=dict(
                owner=owner_address,
                spender=spender_address,
                value=value,
                deadline=deadline,
                v=v,
                r=r,
                s=s,
            ),
            tx_args=kwargs,
        )

    @classmethod
    def allowance(
        cls,
        ledger_api: EthereumApi,
        contract_address: str,
        owner_address: str,
        spender_address: str,
    ) -> Optional[JSONLike]:
        """Gets the allowance for a spender."""
        contract_instance = cls.get_instance(ledger_api, contract_address)

        return ledger_api.contract_method_call(
            contract_instance=contract_instance,
            method_name="allowance",
            owner=owner_address,
            spender=spender_address,
        )

    @classmethod
    def balance_of(
        cls, ledger_api: EthereumApi, contract_address: str, owner_address: str
    ) -> Optional[JSONLike]:
        """Gets an account's balance."""
        contract_instance = cls.get_instance(ledger_api, contract_address)

        return ledger_api.contract_method_call(
            contract_instance=contract_instance,
            method_name="balanceOf",
            owner=owner_address,
        )

    @classmethod
    def get_transaction_transfer_logs(  # type: ignore  # pylint: disable=too-many-arguments,too-many-locals,unused-argument,arguments-differ
        cls,
        ledger_api: EthereumApi,
        contract_address: str,
        tx_hash: str,
        target_address: Optional[str] = None,
    ) -> JSONLike:
        """
        Get all transfer events derived from a transaction.

        :param ledger_api: the ledger API object
        :param contract_address: the address of the contract
        :param tx_hash: the transaction hash
        :param target_address: optional address to filter tranfer events to just those that affect it
        :return: the verified status
        """
        transfer_logs_data: Optional[JSONLike] = super(
            UniswapV2ERC20Contract, cls
        ).get_transaction_transfer_logs(  # type: ignore
            ledger_api,
            tx_hash,
            target_address,
        )
        transfer_logs: List = []

        if transfer_logs_data:

            transfer_logs = cast(
                List,
                transfer_logs_data["logs"],
            )

            transfer_logs = [
                {
                    "from": log["args"]["from"],
                    "to": log["args"]["to"],
                    "value": log["args"]["value"],
                    "token_address": log["address"],
                }
                for log in transfer_logs
            ]

            if target_address:
                transfer_logs = list(
                    filter(
                        lambda log: target_address in (log["from"], log["to"]),  # type: ignore
                        transfer_logs,
                    )
                )

        return dict(logs=transfer_logs)
